@@ -1,12 +1,12 @@
 /*
-**  Copyright (c) 2004-2007 Sendmail, Inc. and its suppliers.
+**  Copyright (c) 2004-2008 Sendmail, Inc. and its suppliers.
 **	All rights reserved.
 **
-**  $Id: dk-filter.c,v 1.172 2007/05/31 18:59:00 msk Exp $
+**  $Id: dk-filter.c,v 1.180 2008/03/12 14:17:17 msk Exp $
 */
 
 #ifndef lint
-static char dk_filter_c_id[] = "@(#)$Id: dk-filter.c,v 1.172 2007/05/31 18:59:00 msk Exp $";
+static char dk_filter_c_id[] = "@(#)$Id: dk-filter.c,v 1.180 2008/03/12 14:17:17 msk Exp $";
 #endif /* !lint */
 
 /* system includes */
@@ -120,7 +120,6 @@ struct Header
 	struct Header *	hdr_next;
 };
 
-#if _FFR_MULTIPLE_KEYS
 /*
 **  KEYTABLE -- table of keys
 */
@@ -133,7 +132,6 @@ struct keytable
 	unsigned char *	key_data;		/* private key data */
 	struct keytable * key_next;		/* next record */
 };
-#endif /* _FFR_MULTIPLE_KEYS */
 
 /*
 **  MSGCTX -- message context, containing transaction-specific data
@@ -154,9 +152,7 @@ struct msgctx
 	DK *		mctx_dk;		/* DomainKeys handle */
 	struct Header *	mctx_hqhead;		/* header queue head */
 	struct Header *	mctx_hqtail;		/* header queue tail */
-#if _FFR_MULTIPLE_KEYS
 	struct keytable * mctx_key;		/* key information */
-#endif /* _FFR_MULTIPLE_KEYS */
 	char		mctx_hlist[MAXHEADERS];	/* header buffer */
 };
 
@@ -311,10 +307,8 @@ unsigned char *seckey;				/* secret key data */
 char *selector;					/* key selector */
 struct Config config;				/* configuration */
 DK_LIB *libdk;					/* libdk handle */
-#if _FFR_MULTIPLE_KEYS
 struct keytable *keyhead;			/* key list */
 struct keytable *keytail;			/* key list */
-#endif /* _FFR_MULTIPLE_KEYS */
 Peer peerlist;					/* queue of "peers" */
 Peer internal;					/* queue of "internal" hosts */
 Peer exignore;					/* "external ignore" hosts */
@@ -356,6 +350,8 @@ static void *fakepriv;				/* fake private space */
 				} \
 			} while (0)
 
+#define	DKF_EOHMACROS	"i {daemon_name} {auth_type}"
+
 
 
 /*
@@ -423,7 +419,6 @@ dkf_zapkey(void)
 		free(seckey);
 	}
 
-#if _FFR_MULTIPLE_KEYS
 	if (keyhead != NULL)
 	{
 		struct keytable *key;
@@ -434,7 +429,6 @@ dkf_zapkey(void)
 			free(key->key_data);
 		}
 	}
-#endif /* _FFR_MULTIPLE_KEYS */
 }
 
 /*
@@ -609,6 +603,7 @@ dkf_cleanup(SMFICTX *ctx)
 	/* release memory */
 	if (dfc != NULL)
 	{
+		TRYFREE(dfc->mctx_domain);
 		/* TRYFREE(dfc->mctx_jobid); */
 
 		if (dfc->mctx_hqhead != NULL)
@@ -931,7 +926,6 @@ dkf_findheader(msgctx dfc, char *hname, int instance)
 	return NULL;
 }
 
-#if _FFR_MULTIPLE_KEYS
 /*
 **  DKF_LOADKEYS -- load multiple keys
 **
@@ -1006,7 +1000,7 @@ dkf_loadkeys(char *file)
 			char *r;
 			char *end;
 			size_t n;
-			struct keytable *new;
+			struct keytable *newkey;
 			int keyfd;
 			struct stat s;
 			char retmp[BUFRSZ];
@@ -1087,8 +1081,8 @@ dkf_loadkeys(char *file)
 
 			(void) sm_strlcat(retmp, "$", sizeof retmp);
 
-			new = (struct keytable *) malloc(sizeof(struct keytable));
-			if (new == NULL)
+			newkey = (struct keytable *) malloc(sizeof(struct keytable));
+			if (newkey == NULL)
 			{
 				fprintf(stderr, "%s: malloc(): %s\n",
 				        progname, strerror(errno));
@@ -1097,10 +1091,10 @@ dkf_loadkeys(char *file)
 				return -1;
 			}
 
-			new->key_next = NULL;
+			newkey->key_next = NULL;
 			r = strrchr(q, '/');
-			new->key_selector = strdup (r == NULL ? q : r + 1);
-			if (new->key_selector == NULL)
+			newkey->key_selector = strdup (r == NULL ? q : r + 1);
+			if (newkey->key_selector == NULL)
 			{
 				fprintf(stderr, "%s: strdup(): %s\n",
 				        progname, strerror(errno));
@@ -1109,13 +1103,13 @@ dkf_loadkeys(char *file)
 				return -1;
 			}
 
-			status = regcomp(&new->key_re, retmp, REG_EXTENDED);
+			status = regcomp(&newkey->key_re, retmp, REG_EXTENDED);
 			if (status != 0)
 			{
 				char err[BUFRSZ];
 
 				memset(err, '\0', sizeof err);
-				(void) regerror(status, &new->key_re,
+				(void) regerror(status, &newkey->key_re,
 				                err, sizeof err);
 				fprintf(stderr,
 				        "%s: %s: line %d: regcomp(): %s\n",
@@ -1125,8 +1119,8 @@ dkf_loadkeys(char *file)
 				return -1;
 			}
 
-			new->key_data = malloc(s.st_size);
-			if (new->key_data == NULL)
+			newkey->key_data = malloc(s.st_size);
+			if (newkey->key_data == NULL)
 			{
 				fprintf(stderr, "%s: malloc(): %s\n",
 				        progname, strerror(errno));
@@ -1134,9 +1128,9 @@ dkf_loadkeys(char *file)
 				fclose(f);
 				return -1;
 			}
-			new->key_len = s.st_size;
+			newkey->key_len = s.st_size;
 
-			n = read(keyfd, new->key_data, s.st_size);
+			n = read(keyfd, newkey->key_data, s.st_size);
 			if (n < s.st_size)
 			{
 				if (n == -1)
@@ -1159,13 +1153,13 @@ dkf_loadkeys(char *file)
 
 			if (keyhead == NULL)
 			{
-				keyhead = new;
-				keytail = new;
+				keyhead = newkey;
+				keytail = newkey;
 			}
 			else
 			{
-				keytail->key_next = new;
-				keytail = new;
+				keytail->key_next = newkey;
+				keytail = newkey;
 			}
 		}
 	}
@@ -1174,7 +1168,6 @@ dkf_loadkeys(char *file)
 
 	return 0;
 }
-#endif /* _FFR_MULTIPLE_KEYS */
 
 /*
 **  DKF_REPORT -- generate a report on failure if possible
@@ -1355,38 +1348,101 @@ dkf_report(msgctx dfc, char *result, char *reason)
 static sfsistat
 mlfi_negotiate(SMFICTX *ctx,
 	unsigned long f0, unsigned long f1,
-	unsigned long f2 __attribute__((unused)),
-	unsigned long f3 __attribute__((unused)),
+	SM_UNUSED(unsigned long f2),
+	SM_UNUSED(unsigned long f3),
 	unsigned long *pf0, unsigned long *pf1,
 	unsigned long *pf2, unsigned long *pf3)
 {
+	unsigned long reqactions = (SMFIF_ADDHDRS|SMFIF_CHGHDRS);
+	unsigned long wantactions = (SMFIF_SETSYMLIST);
+	unsigned long protosteps = (SMFIP_NOHELO|SMFIP_NORCPT|SMFIP_NOUNKNOWN|
+	                            SMFIP_NODATA);
 	connctx cc;
 
-#ifdef SMFIF_QUARANTINE
-	*pf0 = SMFIF_ADDHDRS|SMFIF_CHGHDRS|SMFIF_QUARANTINE;
-#else /* SMFIF_QUARANTINE */
-	*pf0 = SMFIF_ADDHDRS|SMFIF_CHGHDRS;
-#endif /* SMFIF_QUARANTINE */
+	/* initialize connection context */
+	cc = malloc(sizeof(struct connctx));
+	if (cc != NULL)
+	{
+		memset(cc, '\0', sizeof(struct connctx));
+		smfi_setpriv(ctx, cc);
+	}
 
-	*pf1 = 0;
-#ifdef SMFIP_HDR_LEADSPC
+	/* verify the actions we need are available */
+	if (quarantine)
+		reqactions |= SMFIF_QUARANTINE;
+	if ((f0 & reqactions) != reqactions)
+	{
+		if (dolog)
+		{
+			syslog(LOG_ERR,
+			       "mlfi_negotiate(): required milter action(s) not available (got 0x%lx, need 0x%lx)",
+			       f0, reqactions);
+		}
+
+		return SMFIS_REJECT;
+	}
+
+	/* also try to get some nice features */
+	wantactions = (wantactions & f0);
+
+	/* set the actions we want */
+	*pf0 = (reqactions | wantactions);
+
+	/* disable as many protocol steps we don't need as are available */
+	*pf1 = (protosteps & f1);
+
+# ifdef SMFIP_HDR_LEADSPC
+	/* request preservation of leading spaces if possible */
 	if ((f1 & SMFIP_HDR_LEADSPC) != 0)
 	{
-		cc = malloc(sizeof(struct connctx));
 		if (cc != NULL)
 		{
-			memset(cc, '\0', sizeof(struct connctx));
 			cc->cctx_noleadspc = TRUE;
 			*pf1 |= SMFIP_HDR_LEADSPC;
-
-			smfi_setpriv(ctx, cc);
 		}
 	}
-#endif /* SMFIP_HDR_LEADSPC */
+# endif /* SMFIP_HDR_LEADSPC */
 
 	*pf2 = 0;
 	*pf3 = 0;
 
+	/* request macros if able */
+	if (macros != NULL && (wantactions & SMFIF_SETSYMLIST) != 0)
+	{
+		int c;
+		char macrolist[BUFRSZ];
+
+		memset(macrolist, '\0', sizeof macrolist);
+
+		sm_strlcpy(macrolist, DKF_EOHMACROS, sizeof macrolist);
+
+		for (c = 0; macros[c] != NULL; c++)
+		{
+			if (macrolist[0] != '\0')
+				sm_strlcat(macrolist, " ", sizeof macrolist);
+
+			if (sm_strlcat(macrolist, macros[c],
+			               sizeof macrolist) >= sizeof macrolist)
+			{
+				if (dolog)
+				{
+					syslog(LOG_ERR,
+					       "mlfi_negotiate(): macro list overflow");
+				}
+
+				return SMFIS_REJECT;
+			}
+		}
+
+		if (smfi_setsymlist(ctx, SMFIM_EOH, macrolist) != MI_SUCCESS)
+		{
+			if (dolog)
+				syslog(LOG_ERR, "smfi_setsymlist() failed");
+
+			return SMFIS_REJECT;
+		}
+	}
+
 	return SMFIS_CONTINUE;
 }
 #endif /* SMFI_VERSION >= 0x01000000 */
@@ -1683,6 +1739,7 @@ mlfi_eoh(SMFICTX *ctx)
 	connctx cc;
 	msgctx dfc;
 	char *p;
+	char *domain = NULL;
 	Header from;
 	Header hdr;
 	char addr[MAXADDRESS + 1];
@@ -1766,8 +1823,8 @@ mlfi_eoh(SMFICTX *ctx)
 
 	/* extract the sender's domain */
 	sm_strlcpy(addr, from->hdr_val, sizeof addr);
-	status = rfc2822_mailbox_split(addr, &p, &dfc->mctx_domain);
-	if (status != 0 || p == NULL || dfc->mctx_domain == NULL)
+	status = rfc2822_mailbox_split(addr, &p, &domain);
+	if (status != 0 || p == NULL || domain == NULL)
 	{
 		if (dolog)
 		{
@@ -1788,7 +1845,7 @@ mlfi_eoh(SMFICTX *ctx)
 	originok = FALSE;
 
 	/* is it a domain we sign for? */
-	if (!msgsigned && domains != NULL && dfc->mctx_domain != NULL)
+	if (!msgsigned && domains != NULL && domain != NULL)
 	{
 		int n;
 
@@ -1796,8 +1853,7 @@ mlfi_eoh(SMFICTX *ctx)
 		{
 			for (n = 0; dompats[n] != NULL; n++)
 			{
-				status = regexec(dompats[n],
-				                 dfc->mctx_domain,
+				status = regexec(dompats[n], domain,
 				                 0, NULL, 0);
 				if (status == 0)
 				{
@@ -1810,22 +1866,20 @@ mlfi_eoh(SMFICTX *ctx)
 		{
 			for (n = 0; domains[n] != NULL; n++)
 			{
-				if (strcasecmp(dfc->mctx_domain,
-				               domains[n]) == 0)
+				if (strcasecmp(domain, domains[n]) == 0)
 				{
-					dfc->mctx_domain = domains[n];
+					domain = domains[n];
 					domainok = TRUE;
 					break;
 				}
 			}
 		}
 
-		if (subdomains)
+		if (!domainok && subdomains)
 		{
-			p = strchr(dfc->mctx_domain, '.');
-			for (p = strchr(dfc->mctx_domain, '.');
+			for (p = strchr(domain, '.');
 			     p != NULL && !domainok;
-			     p = strchr(p, '.'))
+			     p = strchr(p + 1, '.'))
 			{
 				p++;
 				if (*p == '\0')
@@ -1836,7 +1890,7 @@ mlfi_eoh(SMFICTX *ctx)
 					for (n = 0; dompats[n] != NULL; n++)
 					{
 						status = regexec(dompats[n],
-						                 dfc->mctx_domain,
+						                 domain,
 						                 0, NULL, 0);
 						if (status == 0)
 						{
@@ -1852,7 +1906,7 @@ mlfi_eoh(SMFICTX *ctx)
 						if (strcasecmp(p,
 						               domains[n]) == 0)
 						{
-							dfc->mctx_domain = domains[n];
+							domain = domains[n];
 							domainok = TRUE;
 							break;
 						}
@@ -1860,9 +1914,10 @@ mlfi_eoh(SMFICTX *ctx)
 				}
 			}
 		}
+
+		dfc->mctx_domain = strdup(domain);
 	}
 
-#if _FFR_MULTIPLE_KEYS
 	/* is it a domain we sign for?  (take two) */
 	if (keyhead != NULL)
 	{
@@ -1906,7 +1961,6 @@ mlfi_eoh(SMFICTX *ctx)
 			domainok = TRUE;
 		}
 	}
-#endif /* _FFR_MULTIPLE_KEYS */
 
 	/* see if it came in on an authorized MSA/MTA connection */
 	if (mtas != NULL)
@@ -2005,7 +2059,6 @@ mlfi_eoh(SMFICTX *ctx)
 		                         &status);
 	}
 	else
-#if _FFR_MULTIPLE_KEYS
 	if (dfc->mctx_key != NULL)
 	{
 		dfc->mctx_dk = dk_sign(libdk, dfc->mctx_jobid, NULL,
@@ -2020,14 +2073,6 @@ mlfi_eoh(SMFICTX *ctx)
 		                      dfc->mctx_canonalg, dfc->mctx_signalg,
 		                      &status);
 	}
-#else /* _FFR_MULTIPLE_KEYS */
-	{
-		dfc->mctx_dk = dk_sign(libdk, dfc->mctx_jobid, NULL,
-		                      (dk_sigkey_t *) seckey,
-		                      dfc->mctx_canonalg, dfc->mctx_signalg,
-		                      &status);
-	}
-#endif /* _FFR_MULTIPLE_KEYS */
 	if (dfc->mctx_dk == NULL && status != DK_STAT_OK)
 		return dkf_libstatus(ctx, "dk_new()", status);
 
@@ -2313,13 +2358,11 @@ mlfi_eom(SMFICTX *ctx)
 		if (dfc->mctx_signing)
 		{
 			int hcnt = 0;
-#if _FFR_MULTIPLE_KEYS
 			char *sel;
 
 			sel = selector;
 			if (dfc->mctx_key != NULL)
 				sel = dfc->mctx_key->key_selector;
-#endif /* _FFR_MULTIPLE_KEYS */
 
 			memset(sig, '\0', sizeof sig);
 			status = dk_getsig(dfc->mctx_dk, sig, sizeof sig);
@@ -2374,12 +2417,7 @@ mlfi_eom(SMFICTX *ctx)
 			         "%sa=%s; s=%s; d=%s; c=%s; q=%s;\n\t%sb=",
 			         cc->cctx_noleadspc ? " " : "",
 			         dkf_signalg(dfc->mctx_signalg),
-#if _FFR_MULTIPLE_KEYS
 			         sel == NULL ? DK_DEFAULT_SELECTOR : sel,
-#else /* _FFR_MULTIPLE_KEYS */
-			         selector == NULL ? DK_DEFAULT_SELECTOR
-			                          : selector,
-#endif /* _FFR_MULTIPLE_KEYS */
 			         dfc->mctx_domain,
 			         dkf_canonalg(dfc->mctx_canonalg),
 			         dkf_queryalg(dfc->mctx_queryalg),
@@ -2459,13 +2497,14 @@ mlfi_eom(SMFICTX *ctx)
 			                      val, sizeof val);
 
 			snprintf(header, sizeof header,
-			         "%s%s %s=%s; domainkeys=%s%s%s%s%s",
+			         "%s%s; domainkeys=%s%s%s%s%s header.%s=%s",
 			         cc->cctx_noleadspc ? " " : "",
-			         hostname, hdr, val, authresult,
+			         hostname, authresult,
 			         comment == NULL ? "" : " (",
 			         comment == NULL ? "" : comment,
 			         comment == NULL ? "" : ")",
-                                 !(dkf & DK_FLAG_TESTING) ? "" : " (testing)");
+			         !(dkf & DK_FLAG_TESTING) ? "" : " (testing)",
+			         hdr, val);
 		}
 		else
 		{
@@ -2636,11 +2675,14 @@ struct smfiDesc smfilter =
 {
 	DKF_PRODUCT,	/* filter name */
 	SMFI_VERSION,	/* version code -- do not change */
+	(SMFIF_ADDHDRS | SMFIF_CHGHDRS |
 #ifdef SMFIF_QUARANTINE
-	(SMFIF_ADDHDRS|SMFIF_CHGHDRS|SMFIF_QUARANTINE),	/* flags */
-#else /* SMFIF_QUARANTINE */
-	(SMFIF_ADDHDRS|SMFIF_CHGHDRS),	/* flags */
+	 SMFIF_QUARANTINE |
 #endif /* SMFIF_QUARANTINE */
+#ifdef SMFIF_SETSYMLIST
+	 SMFIF_SETSYMLIST |
+#endif /* SMFIF_SETSYMLIST */
+	 0),		/* flags */
 	mlfi_connect,	/* connection info filter */
 	NULL,		/* SMTP HELO command filter */
 	mlfi_envfrom,	/* envelope sender filter */
@@ -2658,7 +2700,7 @@ struct smfiDesc smfilter =
 	NULL,		/* DATA */
 #endif
 #if SMFI_VERSION >= 0x01000000
-	mlfi_negotiate,	/* negotiation callback */
+	mlfi_negotiate	/* negotiation callback */
 #endif
 };
 #endif /* !DEBUG */
@@ -2915,9 +2957,7 @@ usage(void)
 	                "-H          \tsign with explicit header lists\n"
 	                "-i ilist    \tfile containing list of internal (signing) hosts\n"
 	                "-I elist    \tfile containing list of external domain clients\n"
-#if _FFR_MULTIPLE_KEYS
 	                "-k          \tload a key set instead of a single key\n"
-#endif /* _FFR_MULTIPLE_KEYS */
 	                "-l          \tlog activity to system log\n"
 	                "-m mtalist  \tMTA daemon names for which to sign\n"
 	                "-M macrolist\tMTA macros which enable signing\n"
@@ -2950,9 +2990,7 @@ main(int argc, char **argv)
 	bool autorestart = FALSE;
 	bool gotp = FALSE;
 	bool dofork = TRUE;
-#if _FFR_MULTIPLE_KEYS
 	bool multikey = FALSE;
-#endif /* _FFR_MULTIPLE_KEYS */
 	int c;
 	int status;
 #ifndef DEBUG
@@ -2997,10 +3035,8 @@ main(int argc, char **argv)
 	flushheaders = FALSE;
 #endif /* _FFR_FLUSH_HEADERS */
 	send_reports = FALSE;
-#if _FFR_MULTIPLE_KEYS
 	keyhead = NULL;
 	keytail = NULL;
-#endif /* _FFR_MULTIPLE_KEYS */
 	no_i_whine = TRUE;
 	selector = NULL;
 	seckey = NULL;
@@ -3102,11 +3138,9 @@ main(int argc, char **argv)
 			elist = optarg;
 			break;
 
-#if _FFR_MULTIPLE_KEYS
 		  case 'k':
 			multikey = TRUE;
 			break;
-#endif /* _FFR_MULTIPLE_KEYS*/
 
 		  case 'l':
 #ifndef DEBUG
@@ -3702,7 +3736,6 @@ main(int argc, char **argv)
 	}
 
 	/* load the secret key, if one was specified */
-#if _FFR_MULTIPLE_KEYS
 	if (keyfile != NULL && multikey)
 	{
 		status = dkf_loadkeys(keyfile);
@@ -3710,9 +3743,7 @@ main(int argc, char **argv)
 		if (status != 0)
 			return EX_UNAVAILABLE;
 	}
-	else
-#endif /* _FFR_MULTIPLE_KEYS */
-	if (keyfile != NULL)
+	else if (keyfile != NULL)
 	{
 		int fd;
 		size_t rlen;
@@ -4080,20 +4111,23 @@ main(int argc, char **argv)
 	{
 		int opts;
 
-		opts = DK_OPTS_HDRLIST;
+		opts = DK_LIBFLAGS_HDRLIST;
 
-		(void) dk_options(libdk, DK_OP_SETOPT, &opts);
+		(void) dk_options(libdk, DK_OP_SETOPT, DK_OPTS_FLAGS, &opts,
+		                  sizeof opts);
 	}
 
 	if (send_reports)
 	{
 		int opts;
 
-		(void) dk_options(libdk, DK_OP_GETOPT, &opts);
+		(void) dk_options(libdk, DK_OP_GETOPT, DK_OPTS_FLAGS, &opts,
+		                  sizeof opts);
 
-		opts |= DK_OPTS_TMPFILES;
+		opts |= DK_LIBFLAGS_TMPFILES;
 
-		(void) dk_options(libdk, DK_OP_SETOPT, &opts);
+		(void) dk_options(libdk, DK_OP_SETOPT, DK_OPTS_FLAGS, &opts,
+		                  sizeof opts);
 	}
 
 	pthread_mutex_init(&popen_lock, NULL);
@@ -4201,6 +4235,8 @@ main(int argc, char **argv)
 
 	dkf_zapkey();
 
+	dk_close(libdk);
+
 	return status;
 #endif /* DEBUG */
 }
