@@ -1,12 +1,12 @@
 /*
-**  Copyright (c) 2004, 2005, 2007 Sendmail, Inc. and its suppliers.
+**  Copyright (c) 2004, 2005, 2007, 2008 Sendmail, Inc. and its suppliers.
 **	All rights reserved.
 **
-**  $Id: util.c,v 1.46 2007/05/31 18:58:42 msk Exp $
+**  $Id: util.c,v 1.48 2008/03/12 14:17:17 msk Exp $
 */
 
 #ifndef lint
-static char util_c_id[] = "@(#)$Id: util.c,v 1.46 2007/05/31 18:58:42 msk Exp $";
+static char util_c_id[] = "@(#)$Id: util.c,v 1.48 2008/03/12 14:17:17 msk Exp $";
 #endif /* !lint */
 
 /* system includes */
@@ -62,10 +62,6 @@ static char *optlist[] =
 	"_FFR_FLUSH_HEADERS",
 #endif /* _FFR_FLUSH_HEADERS */
 
-#if _FFR_MULTIPLE_KEYS
-	"_FFR_MULTIPLE_KEYS",
-#endif /* _FFR_MULTIPLE_KEYS */
-
 #if _FFR_REQUIRED_HEADERS
 	"_FFR_REQUIRED_HEADERS",
 #endif /* _FFR_REQUIRED_HEADERS */
@@ -601,13 +597,13 @@ dkf_load_list(FILE *in, struct Peer **list)
 size_t
 dkf_inet_ntoa(struct in_addr a, char *buf, size_t buflen)
 {
-	in_addr_t addr;
+	uint32_t addr;
 
 	assert(buf != NULL);
 
 	addr = ntohl(a.s_addr);
 
-	return snprintf(buf, buflen, "%d.%d.%d.%d",
+	return snprintf(buf, buflen, "%u.%u.%u.%u",
 	                (addr >> 24), (addr >> 16) & 0xff,
 	                (addr >> 8) & 0xff, addr & 0xff);
 }
