@@ -1,8 +1,8 @@
 /*
-**  Copyright (c) 2004 Sendmail, Inc. and its suppliers.
+**  Copyright (c) 2004, 2005, 2007 Sendmail, Inc. and its suppliers.
 **	All rights reserved.
 **
-**  $Id: util.h,v 1.18 2005/12/02 22:36:58 msk Exp $
+**  $Id: util.h,v 1.19 2007/04/07 23:26:36 msk Exp $
 */
 
 #ifndef _UTIL_H_
@@ -20,7 +20,7 @@
 #include "dk-filter.h"
 
 #ifndef lint
-static char util_h_id[] = "@(#)$Id: util.h,v 1.18 2005/12/02 22:36:58 msk Exp $";
+static char util_h_id[] = "@(#)$Id: util.h,v 1.19 2007/04/07 23:26:36 msk Exp $";
 #endif /* !lint */
 
 /* PROTOTYPES */
@@ -29,10 +29,10 @@ extern bool dkf_checkip(Peer list, struct sockaddr *ip);
 #if POPAUTH
 extern bool dkf_checkpopauth(DB *db, struct sockaddr *ip);
 #endif /* POPAUTH */
-extern void dkf_ffrlist(FILE *where);
 extern size_t dkf_inet_ntoa(struct in_addr a, char *buf, size_t buflen);
 extern bool dkf_load_list(FILE *in, struct Peer **list);
 extern void dkf_lowercase(char *str);
+extern void dkf_optlist(FILE *where);
 extern void dkf_setmaxfd(void);
 extern void dkf_splithdrs(char *buf, size_t buflen);
 extern void dkf_stripbrackets(char *addr);
