@@ -1,10 +1,10 @@
 /*
-**  Copyright (c) 2004-2007 Sendmail, Inc. and its suppliers.
+**  Copyright (c) 2004-2008 Sendmail, Inc. and its suppliers.
 **    All rights reserved.
 */
 
 #ifndef lint
-static char ar_c_id[] = "@(#)$Id: ar.c,v 1.61 2007/04/04 17:07:14 msk Exp $";
+static char ar_c_id[] = "@(#)$Id: ar.c,v 1.74 2008/02/19 19:19:28 msk Exp $";
 #endif /* !lint */
 
 /* OS stuff */
@@ -94,9 +94,16 @@ struct ar_query
 	char			q_name[MAXHOSTNAMELEN + 1];
 };
 
+#ifdef AF_INET6
+typedef struct sockaddr_storage SOCKADDR;
+#else /* AF_INET6 */
+typedef struct sockaddr_in SOCKADDR;
+#endif /* AF_INET6 */
+
 struct ar_libhandle
 {
 	int			ar_nsfd;
+	int			ar_nsfdpf;
 	int			ar_control[2];
 	int			ar_flags;
 	int			ar_nscount;
@@ -111,7 +118,7 @@ struct ar_libhandle
 	pthread_mutex_t		ar_lock;
 	unsigned char *		ar_querybuf;
 	unsigned char *		ar_tcpbuf;
-	struct sockaddr_in *	ar_nsaddrs;
+	SOCKADDR *		ar_nsaddrs;
 	void *			(*ar_malloc) (void *closure, size_t nbytes);
 	void			(*ar_free) (void *closure, void *p);
 	void *			ar_closure;
@@ -139,10 +146,10 @@ struct ar_libhandle
 
 static void *ar_malloc(AR_LIB, size_t);
 static void ar_free(AR_LIB lib, void *ptr);
-static int ar_res_init(void);
+static int ar_res_init(AR_LIB);
 
 /*
-**  INTERNAL GLOBALS
+**  LIBRARY GLOBALS
 */
 
 static struct __res_state ar_res;
@@ -196,6 +203,41 @@ ar_free(AR_LIB lib, void *ptr)
 		free(ptr);
 }
 
+/*
+**  AR_SMASHQUEUE -- smash everything in a list of queue handles
+**
+**  Parameters:
+**  	q -- query at the head of the list to clobber
+**
+**  Return value:
+**  	None.
+**
+**  Notes:
+**  	Very destructive.
+*/
+
+static void
+ar_smashqueue(AR_LIB lib, AR_QUERY q)
+{
+	AR_QUERY cur;
+	AR_QUERY next;
+
+	assert(lib != NULL);
+
+	if (q == NULL)
+		return;
+
+	cur = q;
+	while (cur != NULL)
+	{
+		next = cur->q_next;
+
+		ar_free(lib, cur);
+
+		cur = next;
+	}
+}
+
 /*
 **  AR_TIMELEFT -- given a start time and a duration, see how much is left
 **
@@ -419,31 +461,44 @@ ar_reconnect(AR_LIB lib)
 {
 	int c;
 	int saveerrno;
-	struct sockaddr_in sin;
+	int nsnum;
+	int socklen;
+	struct sockaddr *sa;
 
 	assert(lib != NULL);
 
 	close(lib->ar_nsfd);
 	lib->ar_nsfd = -1;
+	lib->ar_nsfdpf = -1;
 
-	/* grab a socket and try to reconnect */
-	lib->ar_nsfd = socket(AF_INET, SOCK_STREAM, 0);
-	if (lib->ar_nsfd >= 0)
+	/* try to connect to someone */
+	for (c = 0; c < lib->ar_nscount; c++)
 	{
-		int nsnum;
+		nsnum = (c + lib->ar_nsidx) % lib->ar_nscount;
 
-		/* try to connect */
-		for (c = 0; c < lib->ar_nscount; c++)
-		{
-			nsnum = (c + lib->ar_nsidx) % lib->ar_nscount;
+		sa = (struct sockaddr *) &lib->ar_nsaddrs[nsnum];
 
-			memcpy(&sin, &lib->ar_nsaddrs[nsnum],
-			       sizeof(struct sockaddr_in));
+#ifdef AF_INET6
+		if (sa->sa_family == AF_INET6)
+			socklen = sizeof(struct sockaddr_in6);
+		else
+			socklen = sizeof(struct sockaddr_in);
+#else /* AF_INET6 */
+		socklen = sizeof(struct sockaddr_in);
+#endif /* AF_INET6 */
 
-			if (connect(lib->ar_nsfd, (struct sockaddr *)&sin,
-			    sizeof(struct sockaddr_in)) == 0)
-				return TRUE;
-		}
+		lib->ar_nsfd = socket(sa->sa_family, SOCK_STREAM, 0);
+		if (lib->ar_nsfd == -1)
+			continue;
+
+		lib->ar_nsfdpf = sa->sa_family;
+
+		if (connect(lib->ar_nsfd, sa, socklen) == 0)
+			return TRUE;
+
+		close(lib->ar_nsfd);
+		lib->ar_nsfd = -1;
+		lib->ar_nsfdpf = -1;
 	}
 
 	saveerrno = errno;
@@ -640,6 +695,11 @@ ar_sendquery(AR_LIB lib, AR_QUERY query)
 	memcpy(&hdr, lib->ar_querybuf, sizeof hdr);
 	query->q_id = hdr.id;
 
+#ifdef DEBUG
+	printf("*** SEND `%s' class=%d type=%d id=%d time=%d\n", query->q_name,
+	       query->q_class, query->q_type, hdr.id, time(NULL));
+#endif /* DEBUG */
+
 	/* send it */
 	if ((lib->ar_flags & AR_FLAG_USETCP) != 0)
 	{
@@ -657,13 +717,36 @@ ar_sendquery(AR_LIB lib, AR_QUERY query)
 	else
 	{
 		int nsnum;
+		int socklen;
+		struct sockaddr *sa;
 
 		nsnum = query->q_tries % lib->ar_nscount;
 
+		sa = (struct sockaddr *) &lib->ar_nsaddrs[nsnum];
+
+		/* change to the right family if needed */
+		if (sa->sa_family != lib->ar_nsfdpf)
+		{
+			close(lib->ar_nsfd);
+			lib->ar_nsfdpf = -1;
+
+			lib->ar_nsfd = socket(sa->sa_family,
+			                      SOCK_DGRAM, 0);
+			if (lib->ar_nsfd != -1)
+				lib->ar_nsfdpf = sa->sa_family;
+		}
+
+#ifdef AF_INET6
+		if (sa->sa_family == AF_INET6)
+			socklen = sizeof(struct sockaddr_in6);
+		else
+			socklen = sizeof(struct sockaddr_in);
+#else /* AF_INET */
+		socklen = sizeof(struct sockaddr_in);
+#endif /* AF_INET */
+
 		n = sendto(lib->ar_nsfd, lib->ar_querybuf,
-		           lib->ar_writelen, 0,
-		           (struct sockaddr *) &lib->ar_nsaddrs[nsnum],
-		           sizeof(struct sockaddr_in));
+		           lib->ar_writelen, 0, sa, socklen);
 	}
 
 	if (n == (size_t) -1)
@@ -692,6 +775,7 @@ static void *
 ar_dispatcher(void *tp)
 {
 	bool wrote;
+	bool usetimeout;
 	int status;
 	int maxfd;
 	AR_LIB lib;
@@ -733,8 +817,9 @@ ar_dispatcher(void *tp)
 		FD_SET(lib->ar_control[1], &rfds);
 
 		/* determine how long to wait */
-		timeout.tv_sec = INT_MAX;
+		timeout.tv_sec = AR_MAXTIMEOUT;
 		timeout.tv_usec = 0;
+		usetimeout = (lib->ar_queries != NULL);
 		for (q = lib->ar_queries; q != NULL; q = q->q_next)
 		{
 			/* check for absolute timeout */
@@ -763,7 +848,7 @@ ar_dispatcher(void *tp)
 
 		/* XXX -- effect a poll if we knew there was more pending */
 		status = select(maxfd + 1, &rfds, &wfds, NULL,
-		                lib->ar_queries != NULL ? &timeout : NULL);
+		                usetimeout ? &timeout : NULL);
 		if (status == -1)
 		{
 			if (errno == EINTR)
@@ -916,6 +1001,11 @@ ar_dispatcher(void *tp)
 				pthread_mutex_unlock(&q->q_lock);
 			}
 
+#ifdef DEBUG
+			printf("*** RECEIVE id=%d time=%d\n", hdr.id,
+			       time(NULL));
+#endif /* DEBUG */
+
 			/* don't recurse if user buffer is too small */
 			if (q != NULL && r > q->q_buflen)
 				q->q_depth = 0;
@@ -1121,30 +1211,84 @@ ar_dispatcher(void *tp)
 **  	None.
 **
 **  Return value:
-**  	Return value from the resolver library; see res_init(3).
+**  	0 on success, -1 on failure.
 */
 
 static int
-ar_res_init(void)
+ar_res_init(AR_LIB new)
 {
+	size_t bytes;
+	SOCKADDR *sa;
+
+	assert(new != NULL);
+
 	h_errno = NETDB_SUCCESS;
 
 	memset(&ar_res, '\0', sizeof ar_res);
 
-#ifdef AR_RES_MANUAL
-	ar_res_manual(&ar_res);
-#else /* AR_RES_MANUAL */
-# if (defined(__RES) && (__RES <= 19960801))
+	/*
+	**  We'll trust that res_init()/res_ninit() will give us things
+	**  like NS counts and retransmission times, but can't always rely
+	**  on it for the nameservers.
+	*/
+
+#if (defined(__RES) && (__RES <= 19960801))
 	/* old-school (bind4) */
 	res_init();
 	memcpy(&ar_res, &_res, sizeof ar_res);
-# else /* defined(__RES) && (__RES <= 19960801) */
+#else /* defined(__RES) && (__RES <= 19960801) */
 	/* new-school (bind8 and up) */
 	(void) res_ninit(&ar_res);
-# endif /* defined(__RES) && (__RES <= 19960801) */
-#endif /* AR_RES_MANUAL */
+#endif /* defined(__RES) && (__RES <= 19960801) */
 
-	return h_errno;
+	new->ar_nscount = ar_res.nscount;
+	new->ar_retry.tv_sec = ar_res.retrans;
+	new->ar_retry.tv_usec = 0;
+	new->ar_retries = ar_res.retry;
+
+	if (new->ar_nscount == 0)
+		new->ar_nscount = MAXNS;
+
+	bytes = sizeof(SOCKADDR) * new->ar_nscount;
+	if (new->ar_malloc != NULL)
+	{
+		new->ar_nsaddrs = (SOCKADDR *) new->ar_malloc(new->ar_closure,
+		                                              bytes);
+	}
+	else
+	{
+		new->ar_nsaddrs = (SOCKADDR *) malloc(bytes);
+	}
+
+	if (new->ar_nsaddrs == NULL)
+		return -1;
+
+	memset(new->ar_nsaddrs, '\0', sizeof(SOCKADDR) * ar_res.nscount);
+
+#if defined(AR_RES_MANUAL) || defined(AF_INET6)
+	ar_res_parse(&new->ar_nscount, new->ar_nsaddrs,
+	             &new->ar_retries, &new->ar_retry.tv_sec);
+#else /* defined(AR_RES_MANUAL) || defined(AF_INET6) */
+	memcpy(new->ar_nsaddrs, ar_res.nsaddr_list,
+	       sizeof(SOCKADDR) * ar_res.nscount);
+
+	/* an address of 0 (INADDR_ANY) should become INADDR_LOOPBACK */
+	for (c = 0; c < new->ar_nscount; c++)
+	{
+		sa = (SOCKADDR *) new->ar_nsaddrs[c];
+		if (sa->sa_family == AF_INET)
+		{
+			struct sockaddr_in *sin;
+
+			sin = (struct sockaddr_in *) sa;
+
+			if (sin->sin_addr.s_addr == INADDR_ANY)
+				sin->sin_addr.s_addr = htonl(INADDR_LOOPBACK);
+		}
+	}
+#endif /* defined(AR_RES_MANUAL) || defined(AF_INET6) */
+
+	return 0;
 }
 
 /*
@@ -1169,7 +1313,9 @@ ar_init(ar_malloc_t user_malloc, ar_free_t user_free, void *user_closure,
         int flags)
 {
 	int status;
+	int c;
 	AR_LIB new;
+	struct sockaddr *sa;
 
 #define TMP_MALLOC(x)	(user_malloc == NULL ? malloc((x)) \
 			                     : user_malloc(user_closure, ((x))));
@@ -1182,14 +1328,12 @@ ar_init(ar_malloc_t user_malloc, ar_free_t user_free, void *user_closure,
 	if (new == NULL)
 		return NULL;
 
-	ar_res_init();
-
-	new->ar_nscount = ar_res.nscount;
 	new->ar_malloc = user_malloc;
 	new->ar_free = user_free;
 	new->ar_closure = user_closure;
 	new->ar_flags = flags;
 	new->ar_nsfd = -1;
+	new->ar_nsfdpf = -1;
 	new->ar_tcpbuflen = 0;
 	new->ar_tcpbuf = NULL;
 	new->ar_pending = NULL;
@@ -1202,9 +1346,12 @@ ar_init(ar_malloc_t user_malloc, ar_free_t user_free, void *user_closure,
 	new->ar_control[1] = -1;
 	new->ar_nsidx = 0;
 	new->ar_writelen = 0;
-	new->ar_retry.tv_sec = ar_res.retrans;
-	new->ar_retry.tv_usec = 0;
-	new->ar_retries = ar_res.retry;
+
+	if (ar_res_init(new) != 0)
+	{
+		TMP_FREE(new);
+		return NULL;
+	}
 
 	if (socketpair(AF_UNIX, SOCK_STREAM, 0, new->ar_control) != 0)
 	{
@@ -1212,10 +1359,47 @@ ar_init(ar_malloc_t user_malloc, ar_free_t user_free, void *user_closure,
 		return NULL;
 	}
 
-	if ((new->ar_flags & AR_FLAG_USETCP) == 0)
-		new->ar_nsfd = socket(AF_INET, SOCK_DGRAM, 0);
-	else
-		new->ar_nsfd = socket(AF_INET, SOCK_STREAM, 0);
+	/* establish socket; connect if necessary */
+	for (c = 0; c < new->ar_nscount; c++)
+	{
+		sa = (struct sockaddr *) &new->ar_nsaddrs[c];
+
+		if ((new->ar_flags & AR_FLAG_USETCP) == 0)	/* UDP */
+		{
+			new->ar_nsfd = socket(sa->sa_family, SOCK_DGRAM, 0);
+			if (new->ar_nsfd != -1)
+			{
+				new->ar_nsfdpf = sa->sa_family;
+				break;
+			}
+		}
+		else						/* TCP */
+		{
+			int socklen;
+
+			new->ar_nsfd = socket(sa->sa_family, SOCK_STREAM, 0);
+			if (new->ar_nsfd == -1)
+				continue;
+#ifdef AF_INET6
+			if (sa->sa_family == AF_INET6)
+				socklen = sizeof(struct sockaddr_in6);
+			else
+				socklen = sizeof(struct sockaddr_in);
+#else /* AF_INET */
+			socklen = sizeof(struct sockaddr_in);
+#endif /* AF_INET */
+
+			if (connect(new->ar_nsfd, sa, socklen) == 0)
+			{
+				new->ar_nsfdpf = sa->sa_family;
+				break;
+			}
+
+			close(new->ar_nsfd);
+			new->ar_nsfd = -1;
+		}
+	}
+
 	if (new->ar_nsfd == -1)
 	{
 		TMP_CLOSE(new->ar_control[0]);
@@ -1224,47 +1408,6 @@ ar_init(ar_malloc_t user_malloc, ar_free_t user_free, void *user_closure,
 		return NULL;
 	}
 
-	new->ar_nsaddrs = TMP_MALLOC(sizeof(struct sockaddr_in) * ar_res.nscount);
-	if (new->ar_nsaddrs == NULL)
-	{
-		TMP_CLOSE(new->ar_control[0]);
-		TMP_CLOSE(new->ar_control[1]);
-		TMP_CLOSE(new->ar_nsfd);
-		TMP_FREE(new);
-		return NULL;
-	}
-	memcpy(new->ar_nsaddrs, ar_res.nsaddr_list,
-	       sizeof(struct sockaddr_in) * ar_res.nscount);
-
-	if ((new->ar_flags & AR_FLAG_USETCP) != 0)
-	{
-		bool conn = FALSE;
-		int c;
-		struct sockaddr_in sin;
-
-		/* try to connect */
-		for (c = 0; !conn && c < ar_res.nscount; c++)
-		{
-			memcpy(&sin, &ar_res.nsaddr_list[c],
-			       sizeof(struct sockaddr_in));
-
-			if (connect(new->ar_nsfd, (struct sockaddr *)&sin,
-			    sizeof(struct sockaddr_in)) == 0)
-				conn = TRUE;
-		}
-
-		/* no connections made */
-		if (conn == 0)
-		{
-			TMP_CLOSE(new->ar_control[0]);
-			TMP_CLOSE(new->ar_control[1]);
-			TMP_CLOSE(new->ar_nsfd);
-			TMP_FREE(new->ar_nsaddrs);
-			TMP_FREE(new);
-			return NULL;
-		}
-	}
-
 	new->ar_querybuf = TMP_MALLOC(new->ar_querybuflen);
 	if (new->ar_querybuf == NULL)
 	{
@@ -1317,12 +1460,29 @@ ar_shutdown(AR_LIB lib)
 	status = pthread_join(lib->ar_dispatcher, NULL);
 	if (status == 0)
 	{
+		void *closure;
+		void (*user_free)(void *, void *);
+
+		close(lib->ar_nsfd);
 		close(lib->ar_control[1]);
 		pthread_mutex_destroy(&lib->ar_lock);
-		free(lib->ar_nsaddrs);
-		free(lib);
 
-		/* XXX -- ar_free() everything we allocated */
+		ar_smashqueue(lib, lib->ar_pending);
+		ar_smashqueue(lib, lib->ar_queries);
+		ar_smashqueue(lib, lib->ar_recycle);
+
+		if (lib->ar_tcpbuf != NULL)
+			ar_free(lib, lib->ar_tcpbuf);
+		ar_free(lib, lib->ar_querybuf);
+		ar_free(lib, lib->ar_nsaddrs);
+
+		closure = lib->ar_closure;
+		user_free = lib->ar_free;
+
+		if (user_free != NULL)
+			user_free(closure, lib);
+		else
+			free(lib);
 	}
 
 	return status;
@@ -1443,6 +1603,17 @@ ar_addquery(AR_LIB lib, char *name, int class, int type, int depth,
 		prev = *p;
 	}
 
+	/* sanity-check the timeout, if provided */
+	if (timeout != NULL)
+	{
+		if (timeout->tv_sec < 0 || timeout->tv_sec > AR_MAXTIMEOUT ||
+		    timeout->tv_usec < 0 || timeout->tv_usec >= 1000000)
+		{
+			errno = EINVAL;
+			return NULL;
+		}
+	}
+
 	pthread_mutex_lock(&lib->ar_lock);
 
 	if ((lib->ar_flags & AR_FLAG_DEAD) != 0)
