@@ -2,11 +2,11 @@
 **  Copyright (c) 2004-2007 Sendmail, Inc. and its suppliers.
 **    All rights reserved.
 **
-**  $Id: dk.c,v 1.181 2007/05/31 02:05:39 msk Exp $
+**  $Id: dk.c,v 1.186 2008/03/09 15:48:02 msk Exp $
 */
 
 #ifndef lint
-static char dk_c_id[] = "@(#)$Id: dk.c,v 1.181 2007/05/31 02:05:39 msk Exp $";
+static char dk_c_id[] = "@(#)$Id: dk.c,v 1.186 2008/03/09 15:48:02 msk Exp $";
 #endif /* !lint */
 
 /* system includes */
@@ -79,10 +79,12 @@ static char dk_c_id[] = "@(#)$Id: dk.c,v 1.181 2007/05/31 02:05:39 msk Exp $";
 #define	DEFTIMEOUT		5	/* default DNS timeout */
 #define MAXPARAMETER		256	/* biggest parameter we accept */
 
+#ifdef _FFR_HASH_BUFFERING
+# define  DK_HASHBUFSIZE	4096	/* hash buffer size */
+#endif /* _FFR_HASH_BUFFERING */
+
 /* defaults */
 #define DK_DEFAULT_SELECTOR	"main"	/* default selector (filename only) */
-#define DK_DNSNAME		"_domainkey"
-					/* common DNS label */
 #define	DK_TMPDIR		"/var/tmp"
 					/* default dir. for temp files */
 
@@ -100,6 +102,17 @@ static char dk_c_id[] = "@(#)$Id: dk.c,v 1.181 2007/05/31 02:05:39 msk Exp $";
 
 /* prototypes */
 static void dk_error(DK *dk, const char *format, ...);
+static void dk_canon __P((DK *, u_char *, size_t));
+#ifdef _FFR_HASH_BUFFERING
+static void dk_canonbuffer __P((DK *, u_char *, size_t));
+#endif /* _FFR_HASH_BUFFERING */
+
+#ifdef _FFR_HASH_BUFFERING
+# define DK_CANON(x,y,z)	dk_canonbuffer(x,y,z)
+#else /* _FFR_HASH_BUFFERING */
+# define DK_CANON(x,y,z)	dk_canon(x,y,z)
+#endif /* _FFR_HASH_BUFFERING */
+
 
 /* parameter lookup table */
 struct lookup
@@ -150,13 +163,15 @@ static int decoder[256] =
 
 struct dk_lib
 {
-	int		dkl_options;
-	char *		dkl_tmpdir;
+	int		dkl_flags;
+	int		dkl_querymethod;
 	void *		(*dkl_malloc) (void *closure, size_t nbytes);
 	void		(*dkl_free) (void *closure, void *p);
 #if USE_ARLIB
 	AR_LIB		dkl_arlib;
 #endif /* USE_ARLIB */
+	char 		dkl_tmpdir[MAXPATHLEN + 1];
+	u_char		dkl_queryinfo[MAXPATHLEN + 1];
 };
 
 /* other stuff */
@@ -245,17 +260,17 @@ dk_mfree(DK_LIB *libhandle, void *closure, void *p)
 static char *
 dk_strdup(DK_LIB *libhandle, void *closure, char *str)
 {
-	char *new;
+	char *newc;
 	size_t len;
 
 	assert(libhandle != NULL);
 	assert(str != NULL);
 
 	len = strlen(str) + 1;
-	new = dk_malloc(libhandle, closure, len);
-	if (new != NULL)
-		sm_strlcpy(new, str, len);
-	return new;
+	newc = dk_malloc(libhandle, closure, len);
+	if (newc != NULL)
+		sm_strlcpy(newc, str, len);
+	return newc;
 
 }
 
@@ -718,19 +733,114 @@ dk_in_use(DK *dk)
 	return 1;
 }
 
+
 /*
-**  DK_GET_KEY -- acquire the key that will be used to either generate
-**                or verify a signature
+**  DK_GET_KEY_FILE -- retrieve a DK key from a text file (for testing)
 **
 **  Parameters:
 **  	dk -- DK handle
+**  	selector -- selector to retrieve
+**  	domain -- domain from which to retrieve
+**  	buf -- buffer into which to write the result
+**  	buflen -- bytes available at "buf"
+**
+**  Return value:
+**  	A DK_STAT_* constant.
+**
+**  Notes:
+**  	The file opened is defined by the library option DK_OPTS_QUERYINFO
+**  	and must be set prior to use of this function.  Failing to do
+**  	so will cause this function to return DK_STAT_CANTVRFY every time.
+**  	The file should contain lines of the form:
+** 
+**  		<selector>._domainkey.<domain> <space> key-data
+**
+**  	Case matching on the left is case-sensitive, but libdkim already
+**  	wraps the domain name to lowercase.
+*/
+
+static DK_STAT
+dk_get_key_file(DK *dk, char *selector, char *domain, u_char *buf,
+                size_t buflen)
+{
+	FILE *f;
+	u_char *p;
+	u_char *p2;
+	char *path;
+	char name[BUFRSZ + 1];
+
+	assert(dk != NULL);
+
+	path = dk->dk_libhandle->dkl_queryinfo;
+	if (path[0] == '\0')
+	{
+		dk_error(dk, "query file not defined");
+		return DK_STAT_CANTVRFY;
+	}
+
+	f = fopen(path, "r");
+	if (f == NULL)
+	{
+		dk_error(dk, "%s: fopen(): %s", path, strerror(errno));
+		return DK_STAT_CANTVRFY;
+	}
+
+	snprintf(name, sizeof name, "%s.%s.%s", selector, DK_DNSNAME, domain);
+
+	memset(buf, '\0', sizeof buf);
+	while (fgets(buf, BUFRSZ, f) != NULL)
+	{
+		p2 = NULL;
+
+		for (p = buf; *p != '\0'; p++)
+		{
+			if (*p == '\n')
+			{
+				*p = '\0';
+				break;
+			}
+			else if (isascii(*p) && isspace(*p))
+			{
+				*p = '\0';
+				p2 = p + 1;
+			}
+			else if (p2 != NULL)
+			{
+				break;
+			}
+		}
+
+		if (strcasecmp(name, buf) == 0)
+		{
+			sm_strlcpy(buf, p2, buflen);
+			fclose(f);
+			return DK_STAT_OK;
+		}
+	}
+
+	fclose(f);
+
+	return DK_STAT_NOKEY;
+}
+
+/*
+**  DK_GET_KEY_DNS -- acquire from DNS the key that will be used to either
+**                    generate or verify a signature
+**
+**  Parameters:
+**  	dk -- DK handle
+**  	selector -- selector to retrieve
+**  	domain -- domain from which to retrieve
+**  	buf -- buffer into which to write the result
+**  	buflen -- bytes available at "buf"
 **
 **  Return value:
 **  	A DK_STAT.
 */
 
 static DK_STAT
-dk_get_key(DK *dk)
+dk_get_key_dns(DK *dk, char *selector, char *domain, u_char *buf,
+               size_t buflen)
 {
 	bool key = FALSE;
 	int status;
@@ -744,20 +854,17 @@ dk_get_key(DK *dk)
 	int l;
 	int type = -1;
 	int class = -1;
-	int plen = 0;
 	size_t anslen;
-	size_t keyidx;
 #if USE_ARLIB
 	AR_QUERY q;
 #endif /* USE_ARLIB */
 	unsigned char *p;
+	unsigned char *r;
 	unsigned char *cp;
+	unsigned char *eob;
 	unsigned char *eom;
-	char *ssel;
 	char qname[MAXHOSTNAMELEN + 1];
 	unsigned char ansbuf[MAXPACKET];
-	char tag[MAXPARAMETER + 1];
-	char value[MAXPARAMETER + 1];
 #if USE_ARLIB
 	struct timeval timeout;
 #endif /* USE_ARLIB */
@@ -765,62 +872,6 @@ dk_get_key(DK *dk)
 
 	assert(dk != NULL);
 
-	if (dk->dk_mode != DK_MODE_VERIFY)
-	{
-		FILE *f;
-		size_t n;
-		struct stat s;
-		char fn[MAXPATHLEN + 1];
-
-		/* already got one? */
-		if (dk->dk_key != NULL)
-			return DK_STAT_OK;
-
-		ssel = dk_sterilize(dk->dk_selector);
-		if (ssel == NULL)
-		{
-			dk_error(dk, "unsafe selector data");
-			return DK_STAT_SYNTAX;
-		}
-
-		memset(fn, '\0', sizeof fn);
-		snprintf(fn, sizeof fn - 1, DK_PRIVATEKEY,
-		         dk->dk_selector == NULL ? DK_DEFAULT_SELECTOR : ssel);
-		f = fopen(fn, "r");
-		if (f == NULL)
-		{
-			dk_error(dk, "%s: fopen(): %s", fn, strerror(errno));
-			return DK_STAT_INTERNAL;
-		}
-
-		if (fstat(fileno(f), &s) != 0)
-		{
-			dk_error(dk, "%s: fstat(): %s", fn, strerror(errno));
-			return DK_STAT_INTERNAL;
-		}
-
-		dk->dk_key = dk_malloc(dk->dk_libhandle, dk->dk_closure,
-		                       (size_t) s.st_size);
-		if (dk->dk_key == NULL)
-		{
-			dk_error(dk, "unable to allocate %d byte(s)",
-			         s.st_size);
-			return DK_STAT_NORESOURCE;
-		}
-		dk->dk_keylen = s.st_size;
-
-		n = fread(dk->dk_key, (unsigned int) s.st_size, 1, f);
-		if (n != 1)
-		{
-			fclose(f);
-			dk_error(dk, "%s: fread(): %s", fn, strerror(errno));
-			return DK_STAT_INTERNAL;
-		}
-
-		fclose(f);
-		return DK_STAT_OK;
-	}
-
 	snprintf(qname, sizeof qname - 1, "%s.%s.%s", dk->dk_selector,
 	         DK_DNSNAME, dk->dk_domain);
 
@@ -983,45 +1034,145 @@ dk_get_key(DK *dk)
 		return DK_STAT_INTERNAL;
 	}
 
-	dk->dk_b64 = dk_malloc(dk->dk_libhandle, dk->dk_closure, n);
-	if (dk->dk_b64 == NULL)
-	{
-		dk_error(dk, "unable to allocate %d byte(s)", n);
-		return DK_STAT_NORESOURCE;
-	}
-	dk->dk_b64len = n;
+	/* decode the TXT record */
+	eob = buf + buflen - 1;
+	memset(buf, '\0', buflen);
 
-	/* decode the tags, allocate memory, and copy */
-	state = 0;
-	keyidx = 0;
-	for (p = cp, l = 0; p <= eom && n >= 0; p++, n--, l--)
+	for (p = cp, l = 0, r = buf;
+	     p <= eom && n >= 0 && r <= eob;
+	     p++, n--, l--)
 	{
-		if (p == eom || n == 0)
-		{
-			if (state == 3)
-			{
-				state = 4;
-			}
-			else if (state == 1 || state == 2)
-			{
-				dk_error(dk,
-				         "syntax error in key record at `%s'",
-				         qname);
-				return DK_STAT_CANTVRFY;
-			}
-			else
-			{
-				break;
-			}
-		}
-
 		/* if appropriate, consume a length byte */
 		if (p < eom && n > 0 && l == 0)
 		{
 			l = *p + 1;
-			continue;
+		}
+		else
+		{
+			*r = *p;
+			r++;
+		}
+	}
+
+	return DK_STAT_OK;
+}
+
+/*
+**  DK_GET_KEY -- acquire the key that will be used to either
+**                generate or verify a signature
+**
+**  Parameters:
+**  	dk -- DK handle
+**
+**  Return value:
+**  	A DK_STAT.
+*/
+
+static DK_STAT
+dk_get_key(DK *dk)
+{
+	bool key = FALSE;
+	DK_STAT status;
+	int state;
+	size_t keyidx;
+	int plen = 0;
+	char *ssel;
+	unsigned char *p;
+	unsigned char *eom;
+	unsigned char buf[BUFRSZ + 1];
+	char tag[MAXPARAMETER + 1];
+	char qname[MAXHOSTNAMELEN + 1];
+	char value[MAXPARAMETER + 1];
+
+	assert(dk != NULL);
+
+	if (dk->dk_mode != DK_MODE_VERIFY)
+	{
+		FILE *f;
+		size_t n;
+		struct stat s;
+		char fn[MAXPATHLEN + 1];
+
+		/* already got one? */
+		if (dk->dk_key != NULL)
+			return DK_STAT_OK;
+
+		ssel = dk_sterilize(dk->dk_selector);
+		if (ssel == NULL)
+		{
+			dk_error(dk, "unsafe selector data");
+			return DK_STAT_SYNTAX;
 		}
 
+		memset(fn, '\0', sizeof fn);
+		snprintf(fn, sizeof fn - 1, DK_PRIVATEKEY,
+		         dk->dk_selector == NULL ? DK_DEFAULT_SELECTOR : ssel);
+		f = fopen(fn, "r");
+		if (f == NULL)
+		{
+			dk_error(dk, "%s: fopen(): %s", fn, strerror(errno));
+			return DK_STAT_INTERNAL;
+		}
+
+		if (fstat(fileno(f), &s) != 0)
+		{
+			dk_error(dk, "%s: fstat(): %s", fn, strerror(errno));
+			return DK_STAT_INTERNAL;
+		}
+
+		dk->dk_key = dk_malloc(dk->dk_libhandle, dk->dk_closure,
+		                       (size_t) s.st_size);
+		if (dk->dk_key == NULL)
+		{
+			dk_error(dk, "unable to allocate %d byte(s)",
+			         s.st_size);
+			return DK_STAT_NORESOURCE;
+		}
+		dk->dk_keylen = s.st_size;
+
+		n = fread(dk->dk_key, (unsigned int) s.st_size, 1, f);
+		if (n != 1)
+		{
+			fclose(f);
+			dk_error(dk, "%s: fread(): %s", fn, strerror(errno));
+			return DK_STAT_INTERNAL;
+		}
+
+		fclose(f);
+		return DK_STAT_OK;
+	}
+
+	switch (dk->dk_querymethod)
+	{
+	  case DK_QUERY_DNS:
+		status = dk_get_key_dns(dk, dk->dk_selector,
+		                        dk->dk_domain, buf, sizeof buf);
+		if (status != DK_STAT_OK)
+			return status;
+
+		break;
+
+	  case DK_QUERY_FILE:
+		status = dk_get_key_file(dk, dk->dk_selector,
+		                         dk->dk_domain, buf, sizeof buf);
+		if (status != DK_STAT_OK)
+			return status;
+
+		break;
+
+	  default:
+		assert(0);
+	}
+
+	snprintf(qname, sizeof qname - 1, "%s.%s.%s", dk->dk_selector,
+	         DK_DNSNAME, dk->dk_domain);
+
+	/* decode the tags, allocate memory, and copy */
+	eom = buf + strlen(buf);
+	state = 0;
+	keyidx = 0;
+	for (p = buf; p <= eom; p++)
+	{
 		switch (state)
 		{
 		  case 0:				/* before tag */
@@ -1056,10 +1207,25 @@ dk_get_key(DK *dk)
 			plen = 0;
 			state = 3;
 			key = (strcasecmp(tag, "p") == 0);
+			if (dk->dk_b64 == NULL && key)
+			{
+				dk->dk_b64len = strlen(p + 1);
+				dk->dk_b64 = dk_malloc(dk->dk_libhandle,
+				                       dk->dk_closure,
+				                       dk->dk_b64len);
+				if (dk->dk_b64 == NULL)
+				{
+					dk_error(dk,
+					         "unable to allocate %d byte(s)",
+					         dk->dk_b64len);
+
+					return DK_STAT_NORESOURCE;
+				}
+			}
 			/* FALLTHROUGH */
 
 		  case 3:				/* in value */
-			if (*p == ';')
+			if (*p == ';' || *p == '\0')
 			{
 				state = 4;
 				/* FALLTHROUGH */
@@ -1114,8 +1280,17 @@ dk_get_key(DK *dk)
 				dk->dk_b64len = keyidx;
 
 				/* empty == revoked */
-				if (value[0] == '\0')
+				if (dk->dk_b64[0] == '\0')
+				{
 					dk->dk_revoked = TRUE;
+				}
+				else if (!dk_decode(dk, DK_DECODE_KEY))
+				{
+					dk_error(dk,
+					         "can't decode key at `%s'",
+					         qname);
+					return DK_STAT_CANTVRFY;
+				}
 			}
 			else if (strcasecmp(tag, "n") == 0)
 			{
@@ -1133,13 +1308,6 @@ dk_get_key(DK *dk)
 
 	/* XXX -- verify valid and required values here instead of dk_eoh()? */
 
-	/* base64-decode the key into dk_key */
-	if (!dk_decode(dk, DK_DECODE_KEY))
-		return DK_STAT_NORESOURCE;
-
-	dk_mfree(dk->dk_libhandle, dk->dk_closure, dk->dk_b64);
-	dk->dk_b64 = NULL;
-
 	return DK_STAT_OK;
 }
 
@@ -1558,38 +1726,160 @@ static DK *
 dk_new(DK_LIB *libhandle, const char *id, void *memclosure,
        dk_canon_t canon_alg, dk_alg_t sign_alg, DK_STAT *statp)
 {
-	DK *new;
+	DK *newdk;
 
 	/* allocate the handle */
-	new = dk_malloc(libhandle, memclosure, sizeof(struct dk));
-	if (new == NULL)
+	newdk = dk_malloc(libhandle, memclosure, sizeof(struct dk));
+	if (newdk == NULL)
 	{
 		*statp = DK_STAT_NORESOURCE;
 		return NULL;
 	}
 
 	/* populate defaults */
-	memset(new, '\0', sizeof(struct dk));
-	new->dk_id = id;
-	new->dk_signalg = (sign_alg == -1 ? DK_SIGN_DEFAULT
+	memset(newdk, '\0', sizeof(struct dk));
+	newdk->dk_id = id;
+	newdk->dk_signalg = (sign_alg == -1 ? DK_SIGN_DEFAULT
 	                                  : sign_alg);
-	new->dk_canonalg = (canon_alg == -1 ? DK_CANON_DEFAULT
+	newdk->dk_canonalg = (canon_alg == -1 ? DK_CANON_DEFAULT
 	                                    : canon_alg);
-	new->dk_querymethod = DK_QUERY_DEFAULT;
-	new->dk_crlf = DK_CRLF_UNKNOWN;
-	new->dk_mode = DK_MODE_UNKNOWN;
-	new->dk_state = DK_STATE_INIT;
-	new->dk_closure = memclosure;
-	new->dk_libhandle = libhandle;
-	new->dk_tmpdir = libhandle->dkl_tmpdir;
-	new->dk_timeout = DEFTIMEOUT;
-	if ((libhandle->dkl_options & DK_OPTS_HDRLIST) != 0)
-		new->dk_flags |= DK_FLAG_HDRLIST;
+	newdk->dk_querymethod = DK_QUERY_UNKNOWN;
+	newdk->dk_crlf = DK_CRLF_UNKNOWN;
+	newdk->dk_mode = DK_MODE_UNKNOWN;
+	newdk->dk_state = DK_STATE_INIT;
+	newdk->dk_closure = memclosure;
+	newdk->dk_libhandle = libhandle;
+	newdk->dk_tmpdir = libhandle->dkl_tmpdir;
+	newdk->dk_timeout = DEFTIMEOUT;
+	if ((libhandle->dkl_flags & DK_LIBFLAGS_HDRLIST) != 0)
+		newdk->dk_flags |= DK_FLAG_HDRLIST;
 
 	*statp = DK_STAT_OK;
-	return new;
+	return newdk;
 }
 
+/*
+**  DK_CANONINIT -- initialize canonicalization
+**
+**  Parameters:
+**  	dk -- DK handle
+*/
+
+static DK_STAT
+dk_canoninit(DK *dk)
+{
+	struct dk_sha1 *sha1;
+	int saveerr;
+
+	assert(dk != NULL);
+
+#ifdef _FFR_HASH_BUFFERING
+	dk->dk_hashbuf = dk_malloc(dk->dk_libhandle, dk->dk_closure,
+	                           DK_HASHBUFSIZE);
+	if (dk->dk_hashbuf == NULL)
+	{
+		dk_error(dk, "unable to allocate %d byte(s)",
+		           DK_HASHBUFSIZE);
+		return DK_STAT_NORESOURCE;
+	}
+	dk->dk_hashbufsize = DK_HASHBUFSIZE;
+	dk->dk_hashbuflen = 0;
+#endif /* _FFR_HASH_BUFFERING */
+
+	sha1 = dk_malloc(dk->dk_libhandle, dk->dk_closure,
+							sizeof(struct dk_sha1));
+
+	if (sha1 == NULL)
+	{
+		dk_error(dk, "unable to allocate %d byte(s)",
+		         sizeof(struct dk_sha1));
+		return DK_STAT_NORESOURCE;
+	}
+	memset(sha1, '\0', sizeof(struct dk_sha1));
+	sha1->sha1_tmpfd = -1;
+	dk->dk_signinfo = (void *) sha1;
+
+	SHA1_Init(&sha1->sha1_sha1);
+
+	if (DK_DEBUG('c') ||
+	    (dk->dk_libhandle->dkl_flags & DK_LIBFLAGS_TMPFILES) != 0)
+	{
+		/* cache the stored stuff for later processing */
+		snprintf(sha1->sha1_tmppath, sizeof sha1->sha1_tmppath - 1,
+		         "%s/dk.%d.XXXXXX", dk->dk_tmpdir, getpid());
+		sha1->sha1_tmpfd = mkstemp(sha1->sha1_tmppath);
+		saveerr = errno;
+		if (sha1->sha1_tmpfd == -1)
+		{
+			/* looks like this is a bug on OpenBSD */
+			(void) unlink(sha1->sha1_tmppath);
+
+			errno = saveerr;
+
+			dk_error(dk, "unable to create temporary file at %s",
+			         sha1->sha1_tmppath);
+			return DK_STAT_NORESOURCE;
+		}
+		sha1->sha1_tmpbio = BIO_new_fd(sha1->sha1_tmpfd, 1);
+
+		if (!DK_DEBUG('c') &&
+	    	    (dk->dk_libhandle->dkl_flags & DK_LIBFLAGS_KEEPFILES) == 0)
+			(void) unlink(sha1->sha1_tmppath);
+	}
+
+	return DK_STAT_OK;
+}
+
+#ifdef _FFR_HASH_BUFFERING
+/*
+**  DK_CANONBUFFER -- buffer for dk_canon()
+**
+**  Parameters:
+**  	dk -- DK handle
+**  	buf -- buffer containing canonicalized data
+**  	buflen -- number of bytes to consume
+**
+**  Return value:
+**  	None.
+*/
+
+static void
+dk_canonbuffer(DK *dk, u_char *buf, size_t buflen)
+{
+	assert(dk != NULL);
+
+	/* NULL buffer or 0 length means flush */
+	if ((buf == NULL || buflen == 0) && dk->dk_hashbuflen > 0)
+	{
+		dk_canon(dk, dk->dk_hashbuf, dk->dk_hashbuflen);
+		dk->dk_hashbuflen = 0;
+		return;
+	}
+
+	/* not enough buffer space; write the buffer out */
+	if (dk->dk_hashbuflen + buflen > dk->dk_hashbufsize)
+	{
+		dk_canon(dk, dk->dk_hashbuf, dk->dk_hashbuflen);
+		dk->dk_hashbuflen = 0;
+	}
+
+	/*
+	**  Now, if the input is bigger than the buffer, write it too;
+	**  otherwise cache it.
+	*/
+
+	if (buflen >= dk->dk_hashbufsize)
+	{
+		dk_canon(dk, buf, buflen);
+	}
+	else
+	{
+		memcpy(&dk->dk_hashbuf[dk->dk_hashbuflen], buf, buflen);
+		dk->dk_hashbuflen += buflen;
+	}
+}
+#endif /* _FFR_HASH_BUFFERING */
+
 /*
 **  DK_CANON -- canonicalize some data
 **
@@ -1647,7 +1937,7 @@ dk_verror(DK *dk, const char *format, va_list va)
 {
 	int flen;
 	int saverr;
-	char *new;
+	char *newdk;
 
 	assert(dk != NULL);
 	assert(format != NULL);
@@ -1676,9 +1966,9 @@ dk_verror(DK *dk, const char *format, va_list va)
 
 		if (flen >= dk->dk_errlen)
 		{
-			new = dk_malloc(dk->dk_libhandle, dk->dk_closure,
+			newdk = dk_malloc(dk->dk_libhandle, dk->dk_closure,
 			                flen + 1);
-			if (new == NULL)
+			if (newdk == NULL)
 			{
 				errno = saverr;
 				return;
@@ -1686,7 +1976,7 @@ dk_verror(DK *dk, const char *format, va_list va)
 
 			dk_mfree(dk->dk_libhandle, dk->dk_closure,
 			         dk->dk_error);
-			dk->dk_error = new;
+			dk->dk_error = newdk;
 			dk->dk_errlen = flen + 1;
 		}
 		else
@@ -1736,6 +2026,7 @@ dk_init(void *(*caller_mallocf)(void *closure, size_t nbytes),
         void (*caller_freef)(void *closure, void *p))
 {
 	DK_LIB *libhandle;
+	char *path;
 
 	/* initialize the resolver */
 	(void) res_init();
@@ -1750,14 +2041,17 @@ dk_init(void *(*caller_mallocf)(void *closure, size_t nbytes),
 	if (libhandle == NULL)
 		return NULL;
 
+	memset(libhandle, '\0', sizeof(struct dk_lib));
+
 	libhandle->dkl_malloc = caller_mallocf;
 	libhandle->dkl_free = caller_freef;
+	libhandle->dkl_flags = DK_LIBFLAGS_DEFAULT;
+	libhandle->dkl_querymethod = DK_QUERY_UNKNOWN;
 
-	libhandle->dkl_options = DK_OPTS_DEFAULT;
-
-	libhandle->dkl_tmpdir = getenv("DK_TMPDIR");
-	if (libhandle->dkl_tmpdir == NULL || libhandle->dkl_tmpdir[0] == '\0')
-		libhandle->dkl_tmpdir = DK_TMPDIR;
+	path = getenv("DK_TMPDIR");
+	if (path == NULL)
+		path = DK_TMPDIR;
+	sm_strlcpy(libhandle->dkl_tmpdir, path, sizeof libhandle->dkl_tmpdir);
 
 #if USE_ARLIB
 	libhandle->dkl_arlib = ar_init(NULL, NULL, NULL, 0);
@@ -1773,6 +2067,31 @@ dk_init(void *(*caller_mallocf)(void *closure, size_t nbytes),
 	return libhandle;
 }
 
+/*
+**  DK_CLOSE -- shut down a DK library package
+**
+**  Parameters:
+**  	lib -- library handle to shut down
+**
+**  Return value:
+**  	None.
+*/
+
+void
+dk_close(DK_LIB *lib)
+{
+	assert(lib != NULL);
+
+#ifdef USE_ARLIB
+	if (lib->dkl_arlib != NULL)
+		(void) ar_shutdown(lib->dkl_arlib);
+#endif /* USE_ARLIB */
+
+	free((void *) lib);
+
+	EVP_cleanup();
+}
+
 /*
 **  DK_SIGN -- allocate a handle for use in a signature operation
 */
@@ -1782,35 +2101,35 @@ dk_sign(DK_LIB *dklib, const char *id, void *memclosure,
         const dk_sigkey_t *secretkey, dk_canon_t canonalg, dk_alg_t signalg,
         DK_STAT *statp)
 {
-	DK *new;
+	DK *newdk;
 
 	assert(dklib != NULL);
 	assert(canonalg == DK_CANON_SIMPLE || canonalg == DK_CANON_NOFWS);
 	assert(signalg == DK_SIGN_RSASHA1);
 
-	new = dk_new(dklib, id, memclosure, canonalg, signalg, statp);
+	newdk = dk_new(dklib, id, memclosure, canonalg, signalg, statp);
 
-	if (new != NULL)
+	if (newdk != NULL)
 	{
-		new->dk_mode = DK_MODE_SIGN;
+		newdk->dk_mode = DK_MODE_SIGN;
 
 		if (secretkey != NULL)
 		{
-			new->dk_keylen = strlen((const char *) secretkey);
-			new->dk_key = (unsigned char *) dk_malloc(dklib,
+			newdk->dk_keylen = strlen((const char *) secretkey);
+			newdk->dk_key = (unsigned char *) dk_malloc(dklib,
 			                                          memclosure,
-		                                                  new->dk_keylen + 1);
-			if (new->dk_key == NULL)
+		                                                  newdk->dk_keylen + 1);
+			if (newdk->dk_key == NULL)
 			{
-				dk_free(new);
+				dk_free(newdk);
 				return NULL;
 			}
 
-			memcpy(new->dk_key, secretkey, new->dk_keylen + 1);
+			memcpy(newdk->dk_key, secretkey, newdk->dk_keylen + 1);
 		}
 	}
 
-	return new;
+	return newdk;
 }
 
 /*
@@ -1820,18 +2139,18 @@ dk_sign(DK_LIB *dklib, const char *id, void *memclosure,
 DK *
 dk_verify(DK_LIB *dklib, const char *id, void *memclosure, DK_STAT *statp)
 {
-	DK *new;
+	DK *newdk;
 	assert(dklib != NULL);
 
-	new = dk_new(dklib, id, memclosure, DK_CANON_UNKNOWN,
+	newdk = dk_new(dklib, id, memclosure, DK_CANON_UNKNOWN,
 	             DK_SIGN_UNKNOWN, statp);
-	if (new != NULL)
-		new->dk_mode = DK_MODE_VERIFY;
-	return new;
+	if (newdk != NULL)
+		newdk->dk_mode = DK_MODE_VERIFY;
+	return newdk;
 }
 
 /*
-**  DK_GETERROR -- return any stored error string from within the DKIM
+**  DK_GETERROR -- return any stored error string from within the DK
 **                 context handle
 */
 
@@ -1975,8 +2294,11 @@ dk_header(DK *dk, u_char *hdr, size_t len)
 		char *user;
 		char *domain;
 		u_char *colon;
+		u_char tmphdr[MAXHEADER + 1];
 
-		colon = (u_char *) strchr(hdr, ':');
+		sm_strlcpy(tmphdr, hdr, sizeof tmphdr);
+
+		colon = (u_char *) strchr(tmphdr, ':');
 
 		status = rfc2822_mailbox_split(colon + 1, &user, &domain);
 		if (status != 0 || user == NULL || domain == NULL)
@@ -1984,8 +2306,8 @@ dk_header(DK *dk, u_char *hdr, size_t len)
 			dk_error(dk, "unable to parse sender `%s'", colon + 1);
 			return DK_STAT_SYNTAX;
 		}
-		sm_strlcpy(dk->dk_uhdrn, hdr,
-		        MIN(sizeof dk->dk_uhdrn, (u_int) (colon - hdr) + 1));
+		sm_strlcpy(dk->dk_uhdrn, tmphdr,
+		        MIN(sizeof dk->dk_uhdrn, (u_int) (colon - tmphdr) + 1));
 		dk_lowercase(dk->dk_uhdrn);
 		snprintf(dk->dk_uhdrv, sizeof dk->dk_uhdrv, "%s@%s", user,
 		         domain);
@@ -2042,6 +2364,7 @@ dk_eoh(DK *dk)
 	char *from;
 	char *hdr;
 	unsigned char *p;
+	int status;
 
 	assert(dk != NULL);
 
@@ -2126,7 +2449,6 @@ dk_eoh(DK *dk)
 	/* if we found a signature but the sender doesn't match it, say so */
 	if (hdr != NULL)
 	{
-		int status;
 		size_t len;
 		char *user;
 		char *domain;
@@ -2200,19 +2522,6 @@ dk_eoh(DK *dk)
 		return DK_STAT_SYNTAX;
 	}
 
-	/* initialize data hash */
-	sha1 = dk_malloc(dk->dk_libhandle, dk->dk_closure,
-	                 sizeof(struct dk_sha1));
-	if (sha1 == NULL)
-	{
-		dk_error(dk, "unable to allocate %d byte(s)",
-		         sizeof(struct dk_sha1));
-		return DK_STAT_NORESOURCE;
-	}
-	memset(sha1, '\0', sizeof(struct dk_sha1));
-	sha1->sha1_tmpfd = -1;
-	dk->dk_signinfo = (void *) sha1;
-
 	/* verify granularity */
 	if (dk->dk_gran != NULL && dk->dk_gran[0] != '\0')
 	{
@@ -2223,32 +2532,9 @@ dk_eoh(DK *dk)
 		}
 	}
 
-	SHA1_Init(&sha1->sha1_sha1);
-
-	if (DK_DEBUG('c') ||
-	    (dk->dk_libhandle->dkl_options & DK_OPTS_TMPFILES) != 0)
-	{
-		/* cache the stored stuff for later processing */
-		snprintf(sha1->sha1_tmppath, sizeof sha1->sha1_tmppath - 1,
-		         "%s/dk.%d.XXXXXX", dk->dk_tmpdir, getpid());
-		sha1->sha1_tmpfd = mkstemp(sha1->sha1_tmppath);
-		saveerr = errno;
-		if (sha1->sha1_tmpfd == -1)
-		{
-			/* looks like this is a bug on OpenBSD */
-			(void) unlink(sha1->sha1_tmppath);
-
-			errno = saveerr;
-
-			dk_error(dk, "unable to create temporary file at %s",
-			         sha1->sha1_tmppath);
-			return DK_STAT_NORESOURCE;
-		}
-		sha1->sha1_tmpbio = BIO_new_fd(sha1->sha1_tmpfd, 1);
-
-		if (!DK_DEBUG('c'))
-			(void) unlink(sha1->sha1_tmppath);
-	}
+	status = dk_canoninit(dk);
+	if (status != DK_STAT_OK)
+		return status;
 
 	/* dk_hdrlist -- set of headers to include */
 	/* dk_hdrset -- set of headers on input */
@@ -2413,6 +2699,18 @@ dk_eoh(DK *dk)
 		}
 	}
 
+	/*
+	**  Apply signature defaults.
+	*/
+
+	if (dk->dk_processing)
+	{
+		if (dk->dk_libhandle->dkl_querymethod != DK_QUERY_UNKNOWN)
+			dk->dk_querymethod = dk->dk_libhandle->dkl_querymethod;
+		else if (dk->dk_querymethod == DK_QUERY_UNKNOWN)
+			dk->dk_querymethod = DK_QUERY_DEFAULT;
+	}
+
 	/*
 	**  If some of the supposedly-signed headers were absent,
 	**  return an error.
@@ -2507,11 +2805,11 @@ dk_body(DK *dk, u_char *buf, size_t len)
 						for (c = 0;
 						     c < dk->dk_blanks;
 						     c++)
-							dk_canon(dk, CRLF, 2);
+							DK_CANON(dk, CRLF, 2);
 
 						dk->dk_blanks = 0;
 
-						dk_canon(dk, wrote, wlen + 1);
+						DK_CANON(dk, wrote, wlen + 1);
 
 						wlen = 0;
 						wrote = p + 1;
@@ -2531,12 +2829,12 @@ dk_body(DK *dk, u_char *buf, size_t len)
 						for (c = 0;
 						     c < dk->dk_blanks;
 						     c++)
-							dk_canon(dk, CRLF, 2);
+							DK_CANON(dk, CRLF, 2);
 
 						dk->dk_blanks = 0;
 
-						dk_canon(dk, wrote, wlen);
-						dk_canon(dk, CRLF, 2);
+						DK_CANON(dk, wrote, wlen);
+						DK_CANON(dk, CRLF, 2);
 
 						wlen = 0;
 						wrote = p + 1;
@@ -2553,7 +2851,7 @@ dk_body(DK *dk, u_char *buf, size_t len)
 
 		/* write what's left */
 		if (wlen > 0)
-			dk_canon(dk, wrote, wlen);
+			DK_CANON(dk, wrote, wlen);
 
 		break;
 
@@ -2586,12 +2884,12 @@ dk_body(DK *dk, u_char *buf, size_t len)
 						for (c = 0;
 						     c < dk->dk_blanks;
 						     c++)
-							dk_canon(dk, CRLF, 2);
+							DK_CANON(dk, CRLF, 2);
 
 						dk->dk_blanks = 0;
 
-						dk_canon(dk, lbuf, lidx);
-						dk_canon(dk, CRLF, 2);
+						DK_CANON(dk, lbuf, lidx);
+						DK_CANON(dk, CRLF, 2);
 
 						lidx = 0;
 					}
@@ -2601,7 +2899,7 @@ dk_body(DK *dk, u_char *buf, size_t len)
 			{
 				if (lidx == sizeof lbuf)
 				{
-					dk_canon(dk, lbuf, lidx);
+					DK_CANON(dk, lbuf, lidx);
 					lidx = 0;
 				}
 
@@ -2613,12 +2911,12 @@ dk_body(DK *dk, u_char *buf, size_t len)
 		}
 
 		if (lidx > 0)
-			dk_canon(dk, lbuf, lidx);
+			DK_CANON(dk, lbuf, lidx);
 
 		break;
 
 	  default:
-		dk_canon(dk, buf, len);
+		DK_CANON(dk, buf, len);
 		break;
 	}
 
@@ -2691,6 +2989,10 @@ dk_eom(DK *dk, DK_FLAGS *dkf)
 			return DK_STAT_NORESOURCE;
 		}
 
+#ifdef _FFR_HASH_BUFFERING
+		dk_canonbuffer(dk, NULL, 0);
+#endif /* _FFR_HASH_BUFFERING */
+
 		/* set up the RSA object */
 		sha1->sha1_rsa = EVP_PKEY_get1_RSA(sha1->sha1_pkey);
 		if (sha1->sha1_rsa == NULL)
@@ -2783,6 +3085,10 @@ dk_getsig(DK *dk, u_char *buf, size_t len)
 			return status;
 	}
 
+#ifdef _FFR_HASH_BUFFERING
+	dk_canonbuffer(dk, NULL, 0);
+#endif /* _FFR_HASH_BUFFERING */
+
 	/* load the private key */
 	key = BIO_new_mem_buf(dk->dk_key, dk->dk_keylen);
 	if (key == NULL)
@@ -2974,37 +3280,99 @@ dk_gethdrs(DK *dk, int *hcnt, u_char *buf, size_t buflen)
 }
 
 /*
-**  DK_OPTIONS -- set/get options
+**  DK_OPTIONS -- get or set a library option
 **
 **  Parameters:
-**  	dklib -- DK library handle
-**  	op -- DK_OP_SETOPT to set options, DK_OP_GETOPT to get options
-**  	opts -- pointer to either the new option set, or where to write the
-**  	        current option set
+**  	lib -- DK library handle
+**  	op -- operation to perform
+**  	opt -- option to get/set
+**  	ptr -- pointer to its old/new value
+**  	len -- memory available at "ptr"
 **
 **  Return value:
-**  	A DK_STAT value.
+**  	A DK_STAT constant.
 */
 
 DK_STAT
-dk_options(DK_LIB *dklib, int op, int *opts)
+dk_options(DK_LIB *lib, int op, dk_opts_t opt, void *ptr, size_t len)
 {
-	assert(dklib != NULL);
+	assert(lib != NULL);
 	assert(op == DK_OP_SETOPT || op == DK_OP_GETOPT);
-	assert(opts != NULL);
+	assert(len != 0);
 
-	switch (op)
+	switch (opt)
 	{
-	  case DK_OP_SETOPT:
-		dklib->dkl_options = *opts;
-		break;
+	  case DK_OPTS_TMPDIR:
+		if (ptr == NULL)
+			return DK_STAT_INVALID;
 
-	  case DK_OP_GETOPT:
-		*opts = dklib->dkl_options;
-		break;
+		if (op == DK_OP_GETOPT)
+		{
+			sm_strlcpy((u_char *) ptr,
+			           lib->dkl_tmpdir, len);
+		}
+		else
+		{
+			sm_strlcpy(lib->dkl_tmpdir, (u_char *) ptr,
+			           sizeof lib->dkl_tmpdir);
+		}
+		return DK_STAT_OK;
+
+	  case DK_OPTS_FLAGS:
+		if (ptr == NULL)
+			return DK_STAT_INVALID;
+
+		if (len != sizeof lib->dkl_flags)
+			return DK_STAT_INVALID;
+
+		if (op == DK_OP_GETOPT)
+		{
+			memcpy(ptr, &lib->dkl_flags, len);
+		}
+		else
+		{
+			memcpy(&lib->dkl_flags, ptr, len);
+		}
+		return DK_STAT_OK;
+
+	  case DK_OPTS_QUERYMETHOD:
+		if (ptr == NULL)
+			return DK_STAT_INVALID;
+
+		if (len != sizeof lib->dkl_querymethod)
+			return DK_STAT_INVALID;
+
+		if (op == DK_OP_GETOPT)
+		{
+			memcpy(ptr, &lib->dkl_querymethod, len);
+		}
+		else
+		{
+			memcpy(&lib->dkl_querymethod, ptr, len);
+		}
+		return DK_STAT_OK;
+
+	  case DK_OPTS_QUERYINFO:
+		if (ptr == NULL)
+			return DK_STAT_INVALID;
+
+		if (op == DK_OP_GETOPT)
+		{
+			sm_strlcpy(ptr, lib->dkl_queryinfo, len);
+		}
+		else
+		{
+			sm_strlcpy(lib->dkl_queryinfo, ptr,
+			           sizeof lib->dkl_queryinfo);
+		}
+		return DK_STAT_OK;
+
+	  default:
+		return DK_STAT_INVALID;
 	}
 
-	return DK_STAT_OK;
+	/* to silence -Wall */
+	return DK_STAT_INTERNAL;
 }
 
 /*
@@ -3065,15 +3433,15 @@ dk_getidentity(DK *dk, char *hname, size_t hnamelen,
 */
 
 DK_STAT
-dk_timeout(DK *dk, int new, int *old)
+dk_timeout(DK *dk, int newt, int *old)
 {
 	assert(dk != NULL);
 
 	if (old != NULL)
 		*old = dk->dk_timeout;
 
-	if (new != -1)
-		dk->dk_timeout = new;
+	if (newt != -1)
+		dk->dk_timeout = newt;
 
 	return DK_STAT_OK;
 }
@@ -3134,6 +3502,10 @@ dk_free(DK *dk)
 	CLOBBER(dk->dk_key);
 	CLOBBER(dk->dk_gran);
 
+#ifdef _FFR_HASH_BUFFERING
+	CLOBBER(dk->dk_hashbuf);
+#endif /* _FFR_HASH_BUFFERING */
+
 	dk_mfree(dk->dk_libhandle, dk->dk_closure, dk);
 
 	return DK_STAT_OK;
