@@ -2,14 +2,14 @@
 **  Copyright (c) 2004-2007 Sendmail, Inc. and its suppliers.
 **    All rights reserved.
 ** 
-**  $Id: dk.h,v 1.52 2007/04/10 23:50:21 msk Exp $
+**  $Id: dk.h,v 1.54 2008/03/09 08:47:59 msk Exp $
 */
 
 #ifndef _SMI_DK_H_
 #define _SMI_DK_H_
 
 #ifndef lint
-static char dk_h_id[] = "@(#)$Id: dk.h,v 1.52 2007/04/10 23:50:21 msk Exp $";
+static char dk_h_id[] = "@(#)$Id: dk.h,v 1.54 2008/03/09 08:47:59 msk Exp $";
 #endif /* !lint */
 
 /* system includes */
@@ -27,7 +27,10 @@ static char dk_h_id[] = "@(#)$Id: dk.h,v 1.52 2007/04/10 23:50:21 msk Exp $";
 **  	pp == patch number
 */
 
-#define	DK_LIB_VERSION		0x00050000
+#define	DK_LIB_VERSION		0x01000000
+
+#define DK_DNSNAME		"_domainkey"
+					/* common DNS label */
 
 /* limits, macros, etc. */
 #define MAXADDRESS		256	/* biggest user@host we accept */
@@ -55,6 +58,7 @@ typedef int DK_STAT;
 #define	DK_STAT_NORESOURCE	6	/* resource unavailable */
 #define	DK_STAT_INTERNAL	7	/* internal error */
 #define	DK_STAT_REVOKED		8	/* key found, but revoked */
+#define	DK_STAT_INVALID		9	/* invalid argument */
 
 /*
 **  DK_CANON -- canonicalization method
@@ -83,8 +87,11 @@ typedef int dk_alg_t;
 **  DK_QUERY -- query method
 */
 
+typedef int dk_query_t;
+
 #define DK_QUERY_UNKNOWN	-1	/* unknown method */
 #define DK_QUERY_DNS		0	/* DNS query method (per the draft) */
+#define DK_QUERY_FILE		1	/* text file method (for testing) */
 
 #define DK_QUERY_DEFAULT	DK_QUERY_DNS
 
@@ -114,14 +121,24 @@ typedef int DK_FLAGS;
 **  DK_OPTS -- options
 */
 
+typedef int dk_opt_t;
+
 #define DK_OP_GETOPT		0
 #define	DK_OP_SETOPT		1
 
-#define	DK_OPTS_NONE		0x00
-#define	DK_OPTS_TMPFILES	0x01
-#define	DK_OPTS_HDRLIST		0x02
+typedef int dk_opts_t;
 
-#define	DK_OPTS_DEFAULT		DK_OPTS_NONE
+#define	DK_OPTS_FLAGS		0
+#define	DK_OPTS_TMPDIR		1
+#define	DK_OPTS_QUERYMETHOD	2
+#define	DK_OPTS_QUERYINFO	3
+
+#define	DK_LIBFLAGS_NONE	0x00
+#define	DK_LIBFLAGS_TMPFILES	0x01
+#define	DK_LIBFLAGS_HDRLIST	0x02
+#define	DK_LIBFLAGS_KEEPFILES	0x04
+
+#define	DK_LIBFLAGS_DEFAULT	DK_LIBFLAGS_NONE
 
 /*
 **  DK -- DomainKeys context
@@ -148,7 +165,7 @@ typedef struct dk_lib DK_LIB;
 */
 
 /*
-**  dk_init -- initialize the DomainKeys package
+**  DK_INIT -- initialize the DomainKeys package
 **
 **  Parameters:
 **  	None.
@@ -161,7 +178,19 @@ extern DK_LIB *dk_init __P((void *(*mallocf)(void *closure, size_t nbytes),
                             void (*freef)(void *closure, void *p)));
 
 /*
-**  dk_sign -- make a new DomainKeys context for signing
+**  DK_CLOSE -- shut down the DK package
+**
+**  Parameters:
+**  	lib -- DK_LIB handle to shut down
+**
+**  Return value:
+**  	None.
+*/
+
+extern void dk_close __P((DK_LIB *lib));
+
+/*
+**  DK_SIGN -- make a new DomainKeys context for signing
 **
 **  Parameters:
 **  	libhandle -- library handle, returned by dk_init()
@@ -186,7 +215,7 @@ extern DK *dk_sign __P((DK_LIB *libhandle, const char *id, void *memclosure,
                         dk_alg_t sign_alg, DK_STAT *statp));
 
 /*
-**  dk_verify -- make a new DomainKeys context for verifying
+**  DK_VERIFY -- make a new DomainKeys context for verifying
 **
 **  Parameters:
 **  	libhandle -- library handle, returned by dk_init()
@@ -204,7 +233,7 @@ extern DK *dk_verify __P((DK_LIB *libhandle, const char *id, void *memclosure,
                           DK_STAT *statp));
 
 /*
-**  dk_header -- process a header
+**  DK_HEADER -- process a header
 **
 **  Parameters:
 **  	dk -- a DK handle previously returned by dk_new()
@@ -218,7 +247,7 @@ extern DK *dk_verify __P((DK_LIB *libhandle, const char *id, void *memclosure,
 extern DK_STAT dk_header __P((DK *dk, u_char *hdr, size_t len));
 
 /*
-**  dk_eoh -- identify end of headers
+**  DK_EOH -- identify end of headers
 **
 **  Parameters:
 **  	dk -- a DK handle previously returned by dk_new()
@@ -231,7 +260,7 @@ extern DK_STAT dk_header __P((DK *dk, u_char *hdr, size_t len));
 extern DK_STAT dk_eoh __P((DK *dk));
 
 /*
-**  dk_body -- process a body chunk
+**  DK_BODY -- process a body chunk
 **
 **  Parameters:
 **  	dk -- a DK handle previously returned by dk_new()
@@ -245,7 +274,7 @@ extern DK_STAT dk_eoh __P((DK *dk));
 extern DK_STAT dk_body __P((DK *dk, u_char *buf, size_t len));
 
 /*
-**  dk_eom -- identify end of body
+**  DK_EOM -- identify end of body
 **
 **  Parameters:
 **  	dk -- a DK handle previously returned by dk_new()
@@ -258,7 +287,7 @@ extern DK_STAT dk_body __P((DK *dk, u_char *buf, size_t len));
 extern DK_STAT dk_eom __P((DK *dk, DK_FLAGS *flags));
 
 /*
-**  dk_getsig -- compute and return a signature for a message
+**  DK_GETSIG -- compute and return a signature for a message
 **
 **  Parameters:
 **  	dk -- a DK handle previously returned by dk_new()
@@ -272,7 +301,7 @@ extern DK_STAT dk_eom __P((DK *dk, DK_FLAGS *flags));
 extern DK_STAT dk_getsig __P((DK *dk, u_char *buf, size_t len));
 
 /*
-**  dk_gethdrs -- report which headers were included in calculation of
+**  DK_GETHDRS -- report which headers were included in calculation of
 **                the signature
 **
 **  Parameters:
@@ -288,7 +317,7 @@ extern DK_STAT dk_getsig __P((DK *dk, u_char *buf, size_t len));
 extern DK_STAT dk_gethdrs __P((DK *dk, int *hcnt, u_char *buf, size_t buflen));
 
 /*
-**  dk_reportinfo -- return info needed to generate failure report
+**  DK_REPORTINFO -- return info needed to generate failure report
 **
 **  Parameters:
 **  	dk -- a DK handle previously returned by dk_new()
@@ -307,17 +336,20 @@ extern DK_STAT dk_reportinfo __P((DK *dk, int *fd, char *raddr, size_t rlen));
 **
 **  Parameters:
 **  	dklib -- DK library handle
-**  	new -- new option bits
-**  	old -- old option bits (returned)
+**  	op -- operation (DK_OPT_GET or DK_OPT_SET)
+**  	opt -- which option (a DK_OPTS_* constant)
+**  	ptr -- value (in or out)
+**  	len -- bytes available at "ptr"
 **
 **  Return value:
 **  	A DK_STAT value.
 */
 
-extern DK_STAT dk_options __P((DK_LIB *dklib, int new, int *old));
+extern DK_STAT dk_options __P((DK_LIB *dklib, int op, dk_opts_t opt,
+                               void *ptr, size_t len));
 
 /*
-**  dk_getidentity -- retrieve signer's apparent identity
+**  DK_GETIDENTITY -- retrieve signer's apparent identity
 **
 **  Parameters:
 **  	dk -- a DK handle previously returned by dk_new()
@@ -336,7 +368,7 @@ extern DK_STAT dk_getidentity __P((DK *dk, char *hname, size_t hnamelen,
                                    char *hval, size_t hvallen));
 
 /*
-**  dk_timeout -- set/get the current DNS timeout
+**  DK_TIMEOUT -- set/get the current DNS timeout
 **
 **  Parameters:
 **  	dk -- a DK handle previously returned by dk_new()
@@ -348,10 +380,10 @@ extern DK_STAT dk_getidentity __P((DK *dk, char *hname, size_t hnamelen,
 **  	A DK_STAT value.
 */
 
-extern DK_STAT dk_timeout __P((DK *dk, int new, int *old));
+extern DK_STAT dk_timeout __P((DK *dk, int newt, int *oldt));
 
 /*
-**  dk_free -- release resources associated with a DK handle
+**  DK_FREE -- release resources associated with a DK handle
 **
 **  Parameters:
 **  	dk -- a DK handle previously returned by dk_new()
@@ -376,7 +408,7 @@ extern DK_STAT dk_free __P((DK *dk));
 extern const char *dk_geterror __P((DK *dk));
 
 /*
-**  rfc2822_mailbox_split -- extract the userid and host from a structured
+**  RFC2822_MAILBOX_SPLIT -- extract the userid and host from a structured
 **                           header
 **
 **  Parameters:
