@@ -1,10 +1,10 @@
 /*
-**  Copyright (c) 2004 Sendmail, Inc. and its suppliers.
+**  Copyright (c) 2004, 2006 Sendmail, Inc. and its suppliers.
 **    All rights reserved.
 */
 
 #ifndef lint
-static char util_c_id[] = "@(#)$Id: util.c,v 1.14 2004/12/02 00:36:28 msk Exp $";
+static char util_c_id[] = "@(#)$Id: util.c,v 1.16 2006/06/01 17:55:48 msk Exp $";
 #endif /* !lint */
 
 /* system includes */
@@ -34,7 +34,7 @@ static char util_c_id[] = "@(#)$Id: util.c,v 1.14 2004/12/02 00:36:28 msk Exp $"
 **
 **  Return value:
 **  	str if the string was OK, otherwise a pointer to a location in
-**  	the string after which data was safe, or NULL if no such point
+**  	the string after which data is safe, or NULL if no such point
 **  	could be determined.
 */
 
@@ -47,14 +47,10 @@ dk_sterilize(char *str)
 	assert(str != NULL);
 
 	ret = str;
-	for (p = str; *p != '\0'; p++)
-	{
-		if (*p == '/')
-			ret = p + 1;
 
-		if (*p == '.' && *(p + 1) == '.')
-			ret = p + 2;
-	}
+	p = strrchr(str, '/');
+	if (p != NULL)
+		ret = p + 1;
 
 	return (*ret == '\0' ? NULL : ret);
 }
